/*
	this is a modification by Marco Comini of the original
	
	
	file		 : aci_unify.pl
	author	   : Vitaly	Lagoon	  <lagoon@cs.bgu.ac.il>
	Last	Update : Aug 27	,1997
	Description : Implementation	of ACI-UNIFICATION algorithm
*/

:- module(aci_unify,[unify_atoms/2,unify_sets/2,aci_unify/2]).

%%
%% import some modules
%%
:- use_module(aci_basics,[flatten_aci/2]).
:- use_module(library(lists)).
:- use_module(library(ordsets)).


%%
%% unify_atoms/2
%%
unify_atoms([N|Args1],[N|Args2]) :-
	unify_sets(Args1,Args2).

%%
%% unify_sets/2
%%
unify_sets([],[]).
unify_sets([A|As],[B|Bs]) :-
	aci_unify(A,B),
	unify_sets(As,Bs).

aci_unify(A,B):-
	flatten_aci(A,AF), flatten_aci(B,BF),
	unify_one_pair(AF,BF).

unify_one_pair(A,B)	:-
	ground(A),
	ground(B),
	!,
	A==B.

unify_one_pair(A,B)	:-
	ord_union(A,B,ToCover),
	ord_setproduct(A,B,AxB),
	covering_u_table(AxB,ToCover,CovUTable),
	keysort(CovUTable,CovUTableS),
	table2graph(CovUTableS,UG),
	u_graph_proceed(UG).


%%
%% select_ground/2
%%
select_ground([],[]).
select_ground([X|Xs],O)	:-
	(ground(X) -> O	= [X|GXs] ;	O =	GXs),
	select_ground(Xs,GXs).


%%
%% cover_u_table/3
%%
covering_u_table([],[],[]).

covering_u_table([X-Y|XsYs],ToCover,[X-L,Y-L|UTRest]) :-
	(X==Y -> ord_subtract(ToCover,[X],ToCoverRest),	!
	  ;
	 (X	@< Y ->
	  ord_subtract(ToCover,[X,Y],ToCoverRest)
	   ;
	  ord_subtract(ToCover,[Y,X],ToCoverRest))),
	 covering_u_table(XsYs,ToCoverRest,UTRest).

covering_u_table([_|XsYs],ToCover,UTable) :-
	covering_u_table(XsYs,ToCover,UTable).


%%
%% table2graph/2
%%
table2graph([],[]).
table2graph([V-I],[V-[I]]) :- !.
table2graph([V-I,W-J|Vs],Out) :-
	(V==W ->
	  Out =	[V-[I|Is]|VsClps], Cont	= [V-Is|VsClps]
	  ;
	  Out =	[V-[I]|VsClps],	Cont = VsClps),
	table2graph([W-J|Vs],Cont).

%%
%% u_graph_proceed/1
%%
u_graph_proceed([]).
u_graph_proceed([V-E|G]) :-
	unify_single(V,E),
	u_graph_proceed(G).


%%
%% unify_single/2
%%
unify_single(V,Y) :- var(V), !,	V=Y.				% variable
unify_single(S,Y) :- simple(S),	!, match_list(Y,S).	% simple (not f(...))
unify_single(T,Y) :- T=..[N,Arg],					% only unary functions!
	lower_level(Y,N,YArgs),
	dup_n(YArgs,Arg,Args),
	unify_sets(YArgs,Args).


%%
%% lower_level/3
%%
lower_level([],_,[]).
lower_level([NT|NTs],N,[T|Ts]) :-
	NT=..[N,T],	lower_level(NTs,N,Ts).


%%
%% dup_n/3
%%
dup_n([],_,[]).
dup_n([_|Xs],Y,[Y|Ys]) :-
	dup_n(Xs,Y,Ys).

%%
%% match_list/2
%%
match_list([],_).
match_list([X|Xs],X) :-
	match_list(Xs,X).
