:- module(verifier,
    [verifySuccess/2,verifySuccess/3,
    verifyIOcall/3,verifyIOcall/4,
    verifyIO/3,verifyIO/4,
    runningtimeof/1]).

:- use_module(abstract,[abstract_atom/2]).
:- use_module(aci_unify,[unify_sets/2]).
:- use_module(canonical,[canonical/2]).
:- use_module(builtins,[builtin/2]).
:- ensure_loaded(library(lists)).

:- dynamic(spec_db/2).
:- dynamic(tp_db/3).

%%%%%%%%%%%%%%%%%%%%%
%                   %
%   verifySuccess   %
%                   %
%%%%%%%%%%%%%%%%%%%%%

verifySuccess(C,Spec) :- verifySuccess(0,C,Spec).

verifySuccess(Lev,C,SpecP) :-
	plus_to_list_spec(SpecP,Spec),
    canonicalSpec(Spec,SpecC),
    !,
    assertSpec(succ,SpecC),
    debugexe(Lev,3,'the spec is',seeSpecDb((spec_db(succ,X),numbervars(X,10,_),write(X)))),
    abstract_the_clause(C,(H:-Bs)),
    debug(Lev,3,'abs cl H:-Bs',(H:-Bs)),
    (verify_success(Lev,(H:-Bs)),!;true),
    debugexe(Lev,3,'Tp is',seeSpecDb((tp_db(succ,X,Y),numbervars((X,Y),10,_),write((X/Y))))).
verifySuccess(_,_,_) :- write('Something is wrong in the specification').

plus_to_list_spec([],[]).
plus_to_list_spec([T|Ts],[CT|CTs]):-plus_to_list_atom(T,CT),plus_to_list_spec(Ts,CTs).

plus_to_list_atom(H,H1):-
	H =.. [F|Args],
	plus_to_list_terms(Args,Args1),
	H1 =.. [F|Args1].
	
plus_to_list_terms([],[]).
plus_to_list_terms([T|Ts],[CT|CTs]):-plus_to_list_term(T,CT),plus_to_list_terms(Ts,CTs).

plus_to_list_term(X,[X]):-var(X),!.
plus_to_list_term(As+B,[BL|AsL]):- !,plus_to_list_subterm(B,BL),plus_to_list_term(As,AsL).
plus_to_list_term(A,AL):- plus_to_list_subterm(A,AL).

plus_to_list_subterm(X,[X]):-var(X),!.
plus_to_list_subterm(T,CT):-plus_to_list_atom(T,CT).

canonicalSpec([],[]).
canonicalSpec([A|As],[CA|CAs]) :- canonicalAtom(A,CA), canonicalSpec(As,CAs).

canonicalAtom(A,CA):- A =.. AL, canonical(AL,CAL), CA =.. CAL.

assertSpec(T,[]) :- retractall(spec_db(T,_)),retractall(tp_db(T,_,_)).
assertSpec(T,[X|Xs]) :- assertSpec(T,Xs),assertz(spec_db(T,X)).

abstract_the_clause((H :- Bs), (HA :- BsA)) :-
    !,
    abstract_atom(H,HA),
    abstract_the_body(Bs,BsA).
abstract_the_clause(H,(HA:-true)) :-
    abstract_atom(H,HA).

abstract_the_body((B,Bs),(BA,BsA)) :- 
    abstract_atom(B,BA), !, abstract_the_body(Bs,BsA).
abstract_the_body(B,(BA,true)) :- abstract_atom(B,BA).

verify_success(Lev,(H:-Bs)):-
    debug(Lev,2,'BEGIN verS',(post(Bs) -> post(H))),
    tp_subst(H,Theta),
    debug(Lev,3,'vars of H:-Bs',Theta/(H:-Bs)),
    copy_term((H,Theta),(H1,Theta1)),
    tp_subst_to_spec_subst(Theta1,Phi),
    \+ (
        debug(Lev,2,'before solvebody/verS_Bs',Bs/Theta),
        solveBodySpec(Lev,succ,Bs),
        debug(Lev,3,'after solvebody/verS_Bs',Bs/Theta),
        canonical(Theta,FC),
        canonicalAtom(H,HC),
        debug(Lev,2,'its canonical form is ',FC/HC),
        cond_assert(Lev,succ,FC,HC),
        \+ (
            debug(Lev,1,'BEGIN check in spec',FC/HC),
            check_if_is_in_spec(Lev,succ,FC,Phi,H1),
            debug(Lev,1,'END check in spec',FC/HC)
            ),
        write('Clause may be wrong because success '),
		list_to_plus_atom(HC,HCP),
        numbervars(HCP,20,_),
        write(HCP),
        write(' (of the head) is not in the succ-specification.'),
        nl
        ),
    write('Clause is OK.'),nl,
    debug(Lev,2,'END verS',(post(Bs) -> post(H))).

tp_subst(H,Hflat) :- H =.. Hflat.

% tp_subst(H,[tp_subst/N|Vs]):-vars_of(H,Vs),length(Vs,N).
% 
% vars_of(V,[V]):- var(V), !.
% vars_of(G,[]):- ground(G), !.
% vars_of(T,Vs):-
%     compound(T),
%     !,
%     T=.. [_|Args],
%     vars_of_aux(Args,UVs),
%     sort(UVs,Vs).
%     
% vars_of_aux([],[]).
% vars_of_aux([Arg|Args],Vs):-
%     vars_of(Arg,V1s),
%     vars_of_aux(Args,V2s),
%     append(V1s,V2s,Vs).

tp_subst_to_spec_subst(X,X).

% tp_subst_to_spec_subst([tp_subst/N|Vs],[spec_subst/N|Vs]).
        
solveBodySpec(Lev,T,(B,Bs)):-
    !,
    debug(Lev,4,'searching for B',B),
    selectAtomSpec(Lev,T,B,BP),
    debug(Lev,4,'unifying (B,BP)',B/BP),
    unifyAtoms(B,BP),
    debug(Lev,4,'now (B,Bs) is',(B,Bs)),
    solveBodySpec(Lev,T,Bs).
solveBodySpec(_,_,true):-!.
solveBodySpec(Lev,_,_):-
    debug(Lev,0,'body not ending with true!',error(solveBodySpec)),
    write('* A serious internal error occured').
    
selectAtomSpec(Lev,T,X,XP):-
    pattern(X,XP),
    debug(Lev,4,'before atom is',XP),
    (XP =.. XP1, builtin(T,XP1), ! ; spec_db(T,XP) ),
    debug(Lev,4,'after atom is',XP).
selectAtomSpec(Lev,T,X,error):-
    pattern(X,XP),
    \+ (XP =.. XP1, builtin(T,XP1)),
    \+ spec_db(T,XP),
    debug(Lev,1,'could not select',T-X),
    write('* A serious error occured: there''s no information for predicate '),
    functor(X,P,N),
    write(P/N),
    write(' in the '),
    write(T),
    write('-specification.'),
    nl,
    !,fail.
    
pattern(X,XP):- X=..[N|Args], same_length(Args,ArgsP), XP=..[N|ArgsP].

unifyAtoms(A,B) :- A=..[N|AArgs], B=..[N|BArgs], unify_sets(AArgs,BArgs).

cond_assert(Lev,T,FC,HC) :-
    debug(Lev,4,'trying to assert',T-FC),
    \+ (numbervars((FC,HC),0,_), tp_db(T,HC,FC)),
    debug(Lev,4,'it is not already in database',tp_db(T,HC,FC)),
    assertz(tp_db(T,HC,FC)),
    !,
    debug(Lev,4,'just asserted',tp_db(T,HC,FC)).
cond_assert(Lev,T,FC,HC):-
    debug(Lev,1,'could not assert',tp_db(T,HC,FC)).

check_if_is_in_spec(Lev,T,Theta,Phi,H):-
    debug(Lev,4,'before solvebody/check',Phi/H),
    solveBodySpec(Lev,T,(H,true)),
    debug(Lev,4,'after solvebody/check',Phi/H),
    canonical(Phi,PhiC),
    debug(Lev,3,'now checking type equivalence',Theta=PhiC),
    check_same_type(Lev,Theta,PhiC).
    
check_same_type(Lev,[_|As],[_|Bs]):-
    debug(Lev,4,'check_same_type',first),
    copy_term(As,As1),
    copy_term(Bs,Bs1),
    numbervars(As1,100,_),
    debug(Lev,4,'unifying',As1=Bs1),
    unify_sets(As1,Bs1),
    debug(Lev,3,'type A is a supertype of B',As>=Bs),
    copy_term(As,As2),
    copy_term(Bs,Bs2),
    numbervars(Bs2,100,_),
    debug(Lev,4,'unifying',As2=Bs2),
    unify_sets(As2,Bs2),
    debug(Lev,3,'type B is a supertype of A',As=<Bs),
    !.
check_same_type(Lev,[_|Bs],[_|As]):-
    debug(Lev,2,'check_same_type',second),
    copy_term(As,As1),
    copy_term(Bs,Bs1),
    numbervars(As1,100,_),
    debug(Lev,4,'unifying',As1=Bs1),
    unify_sets(As1,Bs1),
    debug(Lev,2,'type A is a supertype of B',Bs>=As),
    copy_term(As,As2),
    copy_term(Bs,Bs2),
    numbervars(Bs2,100,_),
    debug(Lev,4,'unifying',As2=Bs2),
    unify_sets(As2,Bs2),
    debug(Lev,2,'type B is a supertype of A',Bs=<As),
    debug(Lev,0,'same types because of second branch of check_same_type',error(check_same_type)),
    write('* A serious internal error occured').

list_to_plus_atom(H,H1):-
	H =.. [F|Args],
	list_to_plus_terms(Args,Args1),
	H1 =.. [F|Args1].
	
list_to_plus_terms([],[]).
list_to_plus_terms([T|Ts],[CT|CTs]):-list_to_plus_term(T,CT),list_to_plus_terms(Ts,CTs).

list_to_plus_term([X],CX) :- list_to_plus_subterm(X,CX).
list_to_plus_term([X|Xs],CXs+CX):-list_to_plus_subterm(X,CX),list_to_plus_term(Xs,CXs).

list_to_plus_subterm(X,X):-var(X),!.
list_to_plus_subterm([X],CX):-!,list_to_plus_subterm(X,CX).
list_to_plus_subterm(T,CT):-list_to_plus_atom(T,CT).

%%%%%%%%%%%%%%%%%%%%
%                  %
%   verifyIOcall   %
%                  %
%%%%%%%%%%%%%%%%%%%%

verifyIOcall(C,SpecI,SpecO) :- verifyIOcall(0,C,SpecI,SpecO).

verifyIOcall(Lev,C,SpecPI,SpecPO) :- 
	plus_to_list_spec(SpecPI,SpecI),
    canonicalSpec(SpecI,SpecIC),
	plus_to_list_spec(SpecPO,SpecO),
    canonicalSpec(SpecO,SpecOC),
    !,
    assertSpec(call,SpecIC),
    debugexe(Lev,3,'the PRE spec is',seeSpecDb((spec_db(call,X),numbervars(X,10,_),write(X)))),
    assertSpec(succ,SpecOC),
    debugexe(Lev,3,'the POST spec is',seeSpecDb((spec_db(succ,X),numbervars(X,10,_),write(X)))),
    abstract_the_clause(C,(H:-Bs)),
    debug(Lev,3,'abs cl H:-Bs',(H:-Bs)),
    (verify_IO_call(Lev,H,Bs,true),!;true),
    debugexe(Lev,3,'Tp PRE is',seeSpecDb((tp_db(call,X,Y),numbervars((X,Y),10,_),write((X/Y))))),
    debugexe(Lev,3,'Tp POST is',seeSpecDb((tp_db(succ,X,Y),numbervars((X,Y),10,_),write((X/Y))))).
verifyIOcall(_,_,_,_) :- write('Something is wrong in the specification').

verify_IO_call(Lev,H,(B,Bs),Cs) :- !, verify_IO_call_n(Lev,H,B,Bs,Cs).
verify_IO_call(Lev,H,true,Cs) :- %this is similar to success
    !,
    debug(Lev,2,'BEGIN verIoCs',(pre(H),post(Cs) -> post(H))),
    tp_subst(H,Theta),
    debug(Lev,3,'vars of H in H:-Cs',Theta/(H:-Cs)),
    copy_term((H,Theta),(H1,Theta1)),
    tp_subst_to_spec_subst(Theta1,Phi),
    \+ (
        debug(Lev,2,'before solvebody/verIoCs_H',H/Theta),
        solveBodySpec(Lev,call,(H,true)),
        debug(Lev,3,'after solvebody/verIoCs_H',H/Theta),
        debug(Lev,2,'before solvebody/verIoCs_Cs',Cs/Theta),
        solveBodySpec(Lev,succ,Cs),
        debug(Lev,3,'after solvebody/verIoCs_Cs',Cs/Theta),
        canonical(Theta,FC),
        canonicalAtom(H,HC),
        debug(Lev,2,'its canonical form is',FC/HC),
        cond_assert(Lev,succ,FC,HC),
        \+ (
            debug(Lev,1,'BEGIN check in POST',FC/HC),
            check_if_is_in_spec(Lev,succ,FC,Phi,H1),
            debug(Lev,1,'END check in POST',FC/HC)
            ),
        write('Clause may be wrong because success '),
		list_to_plus_atom(HC,HCP),
        numbervars(HCP,20,_),
        write(HCP),
        write(' (of the head) is not in the succ-specification.'),
        nl
        ),
    debug(Lev,2,'END verIoCs',(pre(H),post(Cs) -> post(H))),
    write('Clause is OK.'),nl.
verify_IO_call(Lev,_,_,_) :- 
    debug(Lev,0,'body not ending with true!',error(verify_IO_call)),
    write('* A serious internal error occured').

verify_IO_call_n(Lev,H,B,Bs,Cs) :- 
    debug(Lev,2,'BEGIN verIoCc',(pre(H),post(Cs) -> pre(B))),
    tp_subst(B,Theta),
    debug(Lev,3,'vars of B in H:-Cs,B,Bs',Theta/(H:-Cs,B,Bs)),
    copy_term((B,Theta),(B1,Theta1)),
    tp_subst_to_spec_subst(Theta1,Phi),
    \+ (
        debug(Lev,2,'before solvebody/verIoCc_H',H/Theta),
        solveBodySpec(Lev,call,(H,true)),
        debug(Lev,3,'after solvebody/verIoCc_H',H/Theta),
        debug(Lev,2,'before solvebody/verIoCc_Cs',Cs/Theta),
        solveBodySpec(Lev,succ,Cs),
        debug(Lev,3,'after solvebody/verIoCc_Cs',Cs/Theta),
        canonical(Theta,FC),
        canonicalAtom(B,BC),
        debug(Lev,2,'its canonical form is',FC/BC),
        cond_assert(Lev,call,FC,BC),
        \+ (
            debug(Lev,1,'BEGIN check in PRE',FC/BC),
            check_if_is_in_spec(Lev,call,FC,Phi,B1),
            debug(Lev,1,'END check in PRE',FC/BC)
            ),
        write('Clause may be wrong because call '),
		list_to_plus_atom(BC,BCP),
        numbervars(BCP,20,_),
        write(BCP),
        write(' (atom number '),
        bodylength(Lev,Cs,N),
        write(N),
        write(' of body) is not in the call-specification.'),
        nl
        ),
    debug(Lev,2,'END verIoCc',(pre(H),post(Cs) -> pre(B))),
    verify_IO_call(Lev,H,Bs,(B,Cs)).
    
bodylength(Lev,(_,Bs),N1):-!,bodylength(Lev,Bs,N),N1 is N+1.
bodylength(_,true,1):-!.
bodylength(Lev,_,_):-
    debug(Lev,0,'body not ending with true!',error(bodylength)),
    write('* A serious internal error occured').

%%%%%%%%%%%%%%%%
%              %
%   verifyIO   %
%              %
%%%%%%%%%%%%%%%%

verifyIO(C,SpecI,SpecO) :- verifyIO(0,C,SpecI,SpecO).

verifyIO(Lev,C,SpecPI,SpecPO) :- 
	plus_to_list_spec(SpecPI,SpecI),
    canonicalSpec(SpecI,SpecIC),
	plus_to_list_spec(SpecPO,SpecO),
    canonicalSpec(SpecO,SpecOC),
    !,
    assertSpec(call,SpecIC),
    debugexe(Lev,3,'the PRE spec is',seeSpecDb((spec_db(call,X),numbervars(X,10,_),write(X)))),
    assertSpec(succ,SpecOC),
    debugexe(Lev,3,'the POST spec is',seeSpecDb((spec_db(succ,X),numbervars(X,10,_),write(X)))),
    abstract_the_clause(C,(H:-Bs)),
    debug(Lev,3,'abs cl H:-Bs',(H:-Bs)),
    (verify_IO(Lev,H,Bs),!;true),
    debugexe(Lev,3,'Tp PRE is',seeSpecDb((tp_db(call,X,Y),numbervars((X,Y),10,_),write((X/Y))))),
    debugexe(Lev,3,'Tp POST is',seeSpecDb((tp_db(succ,X,Y),numbervars((X,Y),10,_),write((X/Y))))).
verifyIO(_,_,_,_) :- write('Something is wrong in the specification').

verify_IO(Lev,H,Cs) :- %this is similar to success
    debug(Lev,2,'BEGIN verIO',(pre(H),somepost(Cs) -> post(H))),
    tp_subst(H,Theta),
    debug(Lev,3,'vars of H in H:-Cs',Theta/(H:-Cs)),
    copy_term((H,Theta),(H1,Theta1)),
    tp_subst_to_spec_subst(Theta1,Phi),
    \+ (
        debug(Lev,2,'before solvebody/verIO_Cs',Cs/Theta),
        solveConditionalBodySpec(Lev,H,Cs),
        debug(Lev,3,'after solvebody/verIO_Cs',Cs/Theta),
        debug(Lev,2,'before solvebody/verIO_H',H/Theta),
        solveBodySpec(Lev,call,(H,true)),
        debug(Lev,3,'after solvebody/verIO_H',H/Theta),
        canonical(Theta,FC),
        canonicalAtom(H,HC),
        debug(Lev,2,'its canonical form is',FC/HC),
        cond_assert(Lev,succ,FC,HC),
        \+ (
            debug(Lev,1,'BEGIN check in POST',FC/HC),
            check_if_is_in_spec(Lev,succ,FC,Phi,H1),
            debug(Lev,1,'END check in POST',FC/HC)
            ),
        write('Clause may be wrong because success '),
		list_to_plus_atom(HC,HCP),
        numbervars(HCP,20,_),
        write(HCP),
        write(' (of the head) is not in the succ-specification.'),
        nl
        ),
    debug(Lev,2,'END verIO',(pre(H),somepost(Cs) -> post(H))),
    write('Clause is OK.'),nl.

solveConditionalBodySpec(Lev,H,(B,Bs)):-
    !,
    solve_if_implies(Lev,H,B,Bs).
solveConditionalBodySpec(_,_,true):-!.
solveConditionalBodySpec(Lev,_,_):-
    debug(Lev,0,'body not ending with true!',error(solveConditionalBodySpec)),
    write('* A serious internal error occured').
    
solve_if_implies(Lev,H,B,Bs):-
    pre_implies_pre(Lev,H,B),
    !,
    solveConditionalBodySpec(Lev,H,Bs),
    selectAtomSpec(Lev,succ,B,BP),
    debug(Lev,4,'unifying (B,BP)',B=BP),
    unifyAtoms(B,BP),
    debug(Lev,4,'now B is',B).
solve_if_implies(Lev,H,B,Bs):-
    debug(Lev,2,'nothing to do for',H/B),
    solveConditionalBodySpec(Lev,H,Bs).

pre_implies_pre(Lev,H,B) :- 
    debug(Lev,2,'BEGIN pre_implies_pre',(pre(H) -> pre(B))),
    tp_subst(B,Theta),
    debug(Lev,3,'vars of B',Theta/B),
    copy_term((B,Theta),(B1,Theta1)),
    tp_subst_to_spec_subst(Theta1,Phi),
    \+ (
        debug(Lev,2,'before solvebody/prepre_H',H/Theta),
        solveBodySpec(Lev,call,(H,true)),
        debug(Lev,3,'after solvebody/prepre_H',H/Theta),
        canonical(Theta,FC),
        canonicalAtom(B,BC),
        debug(Lev,2,'its canonical form is',FC/BC),
        cond_assert(Lev,call,FC,BC),
        \+ (debug(Lev,1,'BEGIN check in PRE',FC/BC),
            check_if_is_in_spec(Lev,call,FC,Phi,B1),
            debug(Lev,1,'END check in PRE',FC/BC))
        ),
    debug(Lev,2,'END pre_implies_pre',yes).
pre_implies_pre(Lev,_,_) :- 
    debug(Lev,2,'END pre_implies_pre',no),
    fail.

%%%%%%%%%%%%%%%%%%%%%
%                   %
%   runningtimeof   %
%                   %
%%%%%%%%%%%%%%%%%%%%%

runningtimeof(G):-
    statistics(runtime,_),          % init runtime statistics
    G,
    statistics(runtime,[_,RTime]),
    format('~n-----~n~0f msec. runtime~n',[RTime]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                          %
%   debugging predicates   %
%                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

seeSpecDb(G) :- nl,write('** begin **'),nl,seeSpecDbAux(G).
seeSpecDbAux(G):- G,nl,fail.
seeSpecDbAux(_) :- write('** end **'),nl,nl.

debug(Lev,N,Str,A):-
    number(Lev),
    number(N),
    Lev >= N,
    !,
    copy_term(A,A1),numbervars(A1,0,_),
    writestars(N),write(' ['),write(Str),write('] = '),write(A1),nl.
debug(_,_,_,_).

debugexe(Lev,N,Str,G):-
    number(Lev),
    number(N),
    Lev >= N,
    !,
    writestars(N),write(' ['),write(Str),write('] = '),nl,G.
debugexe(_,_,_,_).

writestars(0):-!,write('**').
writestars(N):-write('*'),N1 is N-1,writestars(N1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% That's all folks %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
